/*
 * IEEE 1284 Device ID tool
 *
 * Copyright (C) 1999, 2000  Tim Waugh <tim@cyberelk.demon.co.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 */

#include <sys/time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/io.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <getopt.h>
#include "parport.h"
#include "ppdev.h"

#define BUFSIZE 200

enum
{
  Mfg = 1,
  Mdl = 2,
  Cmd = 4
};

static void version (void)
{
  printf ("deviceid 0.2\n");
  exit (0);
}

static void syntax (void)      
{
  printf ("Usage: deviceid [options] PORT\n\n"
	  "  --help               show this message\n"
	  "  --version            display version information and exit\n"
	  "  --base		  use user I/O rather than a parport device\n"
	  "  --mfg                show manufacturer\n"
	  "  --cmd                show command set\n"
	  "  --mdl                show model\n\n"
	  "If none of --mfg, --cmd, or --mdl are specified, output is the unprocessed\n"
	  "device ID of the peripheral (without the length field).\n");
  exit (0);
}

static void reset (unsigned long base)
{
  unsigned long ctr = base + 2;
  struct timeval tv;
  outb_p (0xc, ctr);
  outb_p (0x8, ctr);
  tv.tv_sec = 0;
  tv.tv_usec = 5000;
  select (0, NULL, NULL, NULL, &tv);
  outb_p (0xc, ctr);
  tv.tv_sec = 0;
  tv.tv_usec = 5000;
  select (0, NULL, NULL, NULL, &tv);
}

static void terminate (unsigned base)
{
  unsigned long ctr = base + 2;
  struct timeval tv;
  outb_p ((inb (ctr) & ~2) | 8, ctr);
  tv.tv_sec = 0;
  tv.tv_usec = 1000;
  select (0, NULL, NULL, NULL, &tv);
  outb_p (inb (ctr) | 2, ctr);
  tv.tv_sec = 0;
  tv.tv_usec = 1000;
  select (0, NULL, NULL, NULL, &tv);
  outb_p (inb (ctr) & ~2, ctr);
}

static int negotiate (unsigned long base, unsigned char mode)
{
  struct timeval tv;
  unsigned long data = base, status = base + 1, ctr = base + 2;
  unsigned char val;
  val = inb (status);
  reset (base);
  outb_p ((inb (ctr) & ~(1|2)) | 8, ctr);
  outb_p (mode, data);
  tv.tv_sec = 0;
  tv.tv_usec = 1000;
  select (0, NULL, NULL, NULL, &tv);
  outb_p ((inb (ctr) & ~8) | 2, ctr);
  tv.tv_sec = 0;
  tv.tv_usec = 1000;
  select (0, NULL, NULL, NULL, &tv);
  val = inb (status);
  if ((val & 0x78) != 0x38)
    {
      outb ((inb (ctr) & ~2) | 8, ctr);
      fprintf (stderr, "status was %#02x\n", val);
      return 1;
    }
  outb_p (inb (ctr) | 1, ctr);
  outb_p (inb (ctr) & ~(1|2), ctr);
  val = inb (status);
  tv.tv_sec = 0;
  tv.tv_usec = 1000;
  select (0, NULL, NULL, NULL, &tv);
  val = inb (status);
  return !(val & 0x10);
}

static size_t read_nibble (unsigned long base, char *buffer, size_t len)
{
  unsigned long data = base, status = base + 1, ctr = base + 2;
  unsigned char byte = 0;
  char *buf = buffer;
  struct timeval tv;
  int i;

  len *= 2; /* in nibbles */
  for (i = 0; i < len; i++)
    {
      unsigned char nibble;
      unsigned char val = inb (status);

      if ((i & 1) == 0 && (val & 8))
	{
	  outb (inb (ctr) | 2, ctr);
	  return i/2;
	}

      outb (inb (ctr) | 2, ctr);
      tv.tv_sec = 0;
      tv.tv_usec = 1000;
      select (0, NULL, NULL, NULL, &tv);
      nibble = inb (status) >> 3;
      nibble &= ~8;
      if ((nibble & 0x10) == 0)
	nibble |= 8;
      nibble &= 0xf;
      outb (inb (ctr) & ~2, ctr);
      tv.tv_sec = 0;
      tv.tv_usec = 1000;
      select (0, NULL, NULL, NULL, &tv);
      if (i & 1)
	{
	  byte |= nibble << 4;
	  *buf++ = byte;
	}
      else
	{
	  byte = nibble;
	}
    }

  return i/2;
}

static int parse_deviceid (const char *buffer, size_t len, int fields)
{
  int advertised;
  size_t at, begin = 2;

  advertised = buffer[0] * 256 + buffer[1];
  if (advertised != len)
    fprintf (stderr, "Advertised length: %d; actual length: %d\n",
	     advertised, len);

  for (at = 0; at <= len; at++)
    {
      char ch;

      if (at < len) ch = buffer[at];

      if (at == len || ch == ';')
	{
	  /* End of a field. */
	  int field;
	  size_t length = at - begin;

	  if (!fields)
	    {
	      if (at < len)
		length++;

	      fwrite (buffer + begin, 1, length, stdout);
	    }
	  else
	    {
	      const char *p = buffer + begin;

	      if (fields & Mfg)
		{
		  if (!strncasecmp (p, "manufacturer:", 13)) {
		    fwrite (p + 13, 1, length - 13, stdout);
		    putchar ('\n');
		  } else if (!strncasecmp (p, "mfg:", 4)) {
		    fwrite (p + 4, 1, length - 4, stdout);
		    putchar ('\n');
		  }
		}

	      if (fields & Mdl)
		{
		  if (!strncasecmp (p, "model:", 6)) {
		    fwrite (p + 6, 1, length - 6, stdout);
		    putchar ('\n');
		  } else if (!strncasecmp (p, "mdl:", 4)) {
		    fwrite (p + 4, 1, length - 4, stdout);
		    putchar ('\n');
		  }
		}

	      if (fields & Cmd)
		{
		  if (!strncasecmp (p, "command set:", 12)) {
		    fwrite (p + 12, 1, length - 12, stdout);
		    putchar ('\n');
		  } else if (!strncasecmp (p, "cmd:", 4)) {
		    fwrite (p + 4, 1, length - 4, stdout);
		    putchar ('\n');
		  }
		}
	    }
	  

	  begin = at + 1;
	}
    }

  if (!fields)
    putchar ('\n');

  return 0;
}

int main (int argc, char *argv[])
{
  int fd;
  int ret;
  int fields = 0;
  int compat = IEEE1284_MODE_COMPAT;
  int mode = IEEE1284_MODE_NIBBLE;
  int userio = 0;

  for (;;)
    {
      static struct option long_opts[] =
      {
	{ "help", 0, 0, 'h' },
	{ "version", 0, 0, 'v' },
	{ "mfg", 0, 0, Mfg },
	{ "cmd", 0, 0, Cmd },
	{ "mdl", 0, 0, Mdl },
	{ "base", 0, 0, 'b' },
	{ 0, 0, 0, 0 }
      };
      int c = getopt_long (argc, argv, "b", long_opts, NULL);

      if (c == -1)
	break;

      switch (c)
	{
	case 'h':
	  syntax ();

	case 'v':
	  version ();

	case 'b':
	  userio = 1;
	  break;

	case Mfg:
	case Cmd:
	case Mdl:
	  fields |= c;
	  break;
	}
    }

  if (argc - optind != 1)
    syntax ();


  if (userio)
    {
      char buffer[BUFSIZE];
      size_t len;
      char *end;
      unsigned long base = strtoul (argv[optind], &end, 0);
      if (end == argv[optind])
	{
	  perror (argv[optind]);
	  return fd;
	}

      if (ioperm (base, 3, 1) || ioperm (0x80, 1, 1))
	{
	  fprintf (stderr, "Permission denied.\n");
	  return 1;
	}

      if (negotiate (base, 4))
	{
	  fprintf (stderr, "Couldn't go to nibble mode\n");
	  return 1;
	}

      len = read_nibble (base, buffer, BUFSIZE);
      terminate (base);
      return parse_deviceid (buffer, len, fields);
    }

  fd = open (argv[optind], O_RDONLY | O_NOCTTY);

  if (fd == -1)
    {
      perror (argv[optind]);
      return fd;
    }

  if (ioctl (fd, PPSETMODE, &mode))
    {
      fprintf (stderr, "deviceid: cannot set transfer mode\n");
      close (fd);
      return 1;
    }

  if (ioctl (fd, PPCLAIM))
    {
      fprintf (stderr, "deviceid: cannot claim port\n");
      close (fd);
      return 1;
    }

  if (ioctl (fd, PPNEGOT, &compat))
    {
      perror ("PPNEGOT");
      close (fd);
      return 1;
    }

  mode |= IEEE1284_DEVICEID;
  if (ioctl (fd, PPNEGOT, &mode))
    {
      fprintf (stderr, "deviceid: device not IEEE 1284 compliant\n");
      close (fd);
      return 1;
    }

  {
    char buffer[BUFSIZE];
    ssize_t len = 0;

    if (read (fd, buffer, 2) < 2)
      {
	fprintf (stderr, "deviceid: no length field? (!)\n");
	close (fd);
	return 1;
      }

    len = read (fd, buffer, BUFSIZE);
    ret = parse_deviceid (buffer, len, fields);
  }

  ioctl (fd, PPNEGOT, &compat);
  ioctl (fd, PPRELEASE);
  close (fd);

  return ret;
}
